from django.conf import settings
from django.core.checks import Warning, register, Tags
from .app_settings import app_settings


@register(Tags.compatibility)
def check_settings_combination(app_configs, **kwargs):
    warnings = []

    if app_settings.ENABLE_MIMICRY and not settings.DEBUG:
        warnings.append(
            Warning(
                "'ENABLE_MIMICRY' being true at the same time as you are in production is very inefficient and "
                "not very very secure.",
                hint="Review your settings and adjust as needed.",
                id="mimicry.W001",
            )
        )

    return warnings



from django.contrib.auth import get_user_model
from django.http import HttpRequest, HttpResponsePermanentRedirect, HttpResponseRedirect
from django.shortcuts import redirect
from .forms import UserSwitchForm
from .utils import new_simulated_session, disable_bar, set_hide_bar


UserModel = get_user_model()


# noinspection PyProtectedMember
def switch_user(request: HttpRequest) -> HttpResponsePermanentRedirect | HttpResponseRedirect:
    """
    Switch the current user or simulate a user based on the form data submitted.

    This view supports:
    1. Hiding the simulation bar.
    2. Enabling/disabling user simulation.
    3. Simulating as an unauthenticated user.
    4. Switching to a user based on PK, username, or email.

    If the user is not found by PK, username, or email, the request is redirected to the previous page.
    If the simulation is disabled, the bar is hidden, and any simulated user state is reset.

    Args:
    - request: The request object.

    Returns:
    - HttpResponse: Redirects to the referring page or, if not available, the root URL.
    """
    if request.method == 'POST':
        form: UserSwitchForm = UserSwitchForm(request.POST)
        if form.is_valid():
            # Check if the "hide" checkbox is ticked.
            if form.cleaned_data.get('hide'):
                request: HttpRequest = set_hide_bar(request, True)

            # Check if the simulation is enabled.
            if form.cleaned_data.get('enabled'):
                user_pk: str | None = form.cleaned_data.get('user_pk')
                username: str | None = form.cleaned_data.get('username')
                email: str | None = form.cleaned_data.get('email')
                unauthenticated: bool = form.cleaned_data.get('unauthenticated')

                if unauthenticated:
                    request.simulated_user_pk = 'unauthenticated'
                    request: HttpRequest = new_simulated_session(request)
                elif user_pk:
                    primary_key_name: str = UserModel._meta.pk.name
                    try:
                        user: UserModel = UserModel.objects.get(**{primary_key_name: user_pk})
                        request.simulated_user_pk = getattr(user, primary_key_name)
                        request: HttpRequest = new_simulated_session(request)
                    except UserModel.DoesNotExist:
                        return redirect(request.META.get('HTTP_REFERER', '/'))
                elif username:
                    try:
                        user: UserModel = UserModel.objects.get(username=username)
                        request.simulated_user_pk = getattr(user, UserModel._meta.pk.name)
                        request: HttpRequest = new_simulated_session(request)
                    except UserModel.DoesNotExist:
                        return redirect(request.META.get('HTTP_REFERER', '/'))
                elif email:
                    try:
                        user: UserModel = UserModel.objects.get(email=email)
                        request.simulated_user_pk = getattr(user, UserModel._meta.pk.name)
                        request: HttpRequest = new_simulated_session(request)
                    except UserModel.DoesNotExist:
                        return redirect(request.META.get('HTTP_REFERER', '/'))
                else:
                    return redirect(request.META.get('HTTP_REFERER', '/'))
            else:
                # Disable the simulation bar and reset the simulated user state.
                request: HttpRequest = disable_bar(request)
                if getattr(request, 'simulated_user_pk', None):
                    request.simulated_user_pk = None

    # Redirect to the referring page or the root URL if the referrer is not available.
    return redirect(request.META.get('HTTP_REFERER', '/'))



"""
Custom Template Tags for Private Content Handling in the Mimicry app.

This module provides template tags that protect private content from being displayed during a user simulation.

Attributes:
- `register`: This is a decorator that registers custom tags with Django.
- `private`: A simple tag that replaces content if the user is in a simulation.
- `do_privateblock`: A tag decorator that supports block-level privacy checks.
- `PrivateBlockNode`: A node that represents the block-level privacy tag.

Usage:
In Django templates, you can use the `{% load private_tags %}` directive to import these tags.
Then, use `{% private "content" %}` for single line content protection, or use
`{% privateblock %}...{% endprivateblock %}` for multi-line content protection.
"""

from django import template
from ..app_settings import app_settings


register = template.Library()


@register.simple_tag(takes_context=True)
def private(context, content):
    """
    A simple tag that checks if the user is in a simulation.
    If so, it replaces the provided content with the PRIVATE_CONTENT_REPLACEMENT setting value.
    """
    if context['request'].user != context['request'].real_user:
        return app_settings.PRIVATE_CONTENT_REPLACEMENT
    return content


@register.tag(name="privateblock")
def do_privateblock(parser, token):
    """
    Begins a block-level privacy check. Content inside this block will be replaced if the user is in a simulation.
    Usage:
        {% privateblock %}
            ... protected content ...
        {% endprivateblock %}
    """
    nodelist = parser.parse(('endprivateblock',))
    parser.delete_first_token()
    return PrivateBlockNode(nodelist)


class PrivateBlockNode(template.Node):
    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        """
        Renders the block content and replaces it with PRIVATE_CONTENT_REPLACEMENT if the user is in a simulation.
        """
        output = self.nodelist.render(context)
        if context['request'].user != context['request'].real_user:
            return app_settings.PRIVATE_CONTENT_REPLACEMENT
        return output



from django.contrib import admin
from .models import SimulatedUserSession, SimulatedUserAction


class SimulatedUserSessionAdmin(admin.ModelAdmin):
    """
    Admin configuration for the SimulatedUserSession model.

    Provides:
        - List display with specific fields.
        - Search capability across specific fields.
        - Filtering based on creation and closure dates.
        - Ordering by creation date in descending order.
        - Restriction of permissions to superusers only.
    """

    # Configuration
    list_display: list[str] = [
        'id', 'real_user', 'simulated_user', 'real_session_key', 'simulated_session_key', 'created_at', 'closed_at'
    ]
    search_fields: list[str] = [
        'id', 'real_user', 'simulated_user', 'real_session_key', 'simulated_session_key', 'created_at', 'closed_at'
    ]

    list_filter: list[str] = ['created_at', 'closed_at']

    ordering: list[str] = ['-created_at']

    def has_view_permission(self, request, obj=None) -> bool:
        """Ensure only superusers can view the records."""
        return request.user.is_superuser

    def has_change_permission(self, request, obj=None) -> bool:
        """Ensure only superusers can change the records."""
        return request.user.is_superuser

    def has_add_permission(self, request) -> bool:
        """Ensure only superusers can add new records."""
        return request.user.is_superuser

    def has_delete_permission(self, request, obj=None) -> bool:
        """Ensure only superusers can delete the records."""
        return request.user.is_superuser


class SimulatedUserActionAdmin(admin.ModelAdmin):
    """
    Admin configuration for the SimulatedUserAction model.

    Provides:
        - List display with specific fields.
        - Search capability across specific fields.
        - Filtering based on the timestamp.
        - Ordering by timestamp in descending order.
        - Restriction of permissions to superusers only.
    """

    # Configuration
    list_display: list[str] = ['simulated_user_session', 'action', 'timestamp']
    search_fields: list[str] = ['simulated_user_session', 'action']
    list_filter: list[str] = ['timestamp']
    ordering: list[str] = ['-timestamp']

    def has_view_permission(self, request, obj=None) -> bool:
        """Ensure only superusers can view the records."""
        return request.user.is_superuser

    def has_change_permission(self, request, obj=None) -> bool:
        """Ensure only superusers can change the records."""
        return request.user.is_superuser

    def has_add_permission(self, request) -> bool:
        """Ensure only superusers can add new records."""
        return request.user.is_superuser

    def has_delete_permission(self, request, obj=None) -> bool:
        """Ensure only superusers can delete the records."""
        return request.user.is_superuser


# Registering the models with their respective admin configurations
admin.site.register(SimulatedUserSession, SimulatedUserSessionAdmin)
admin.site.register(SimulatedUserAction, SimulatedUserActionAdmin)



from typing import Type
from django import forms
from django.contrib.auth import get_user_model
from django.db.models import ForeignKey


UserModel = get_user_model()


# noinspection PyProtectedMember
class UserSwitchForm(forms.Form):
    """
    Form for switching or simulating users based on their primary key (PK), username, or email.

    Fields:
    - user_pk: The primary key of the user. Its form field type is determined by the nature of the primary key
               of the custom user model (either a ForeignKey or CharField).
    - username: The username of the user to switch/simulate.
    - email: The email of the user to switch/simulate.
    - unauthenticated: A boolean field indicating if the user wants to simulate as an unauthenticated user.
    - enabled: A boolean field indicating if the user simulation feature should be enabled.
    - hide: A boolean field for deciding whether to hide the simulation UI components, e.g., a control bar.

    Usage:
    This form can be used in views to decide which user to simulate or switch to. It also controls settings like
    whether to show a simulation control bar (through the 'hide' field) and whether the simulation feature is active
    (through the 'enabled' field).
    """
    # Determine the name and form field class of the primary key based on the custom user model's configuration.
    primary_key_name: str = UserModel._meta.pk.name
    primary_key_field_class: Type[forms.ModelChoiceField | forms.CharField] = forms.ModelChoiceField if isinstance(
        UserModel._meta.pk, ForeignKey
    ) else forms.CharField

    user_pk: primary_key_field_class = primary_key_field_class(
        required=False,
        label=primary_key_name.capitalize().replace('_', ' ')
    )
    username: str | None = forms.CharField(required=False, label='Username')
    email: str | None = forms.EmailField(required=False, label='Email')
    unauthenticated: bool = forms.BooleanField(required=False, label='Unauthenticated')
    enabled: bool = forms.BooleanField(required=False, label='Enabled')
    hide: bool = forms.BooleanField(required=False, label='Hide')



"""
Settings Configuration for the Mimicry app.

This module provides default settings for the app and also retrieves overridden settings from the project's main `
settings.py` file if available.

Attributes:
- `MimicrySettings`: A class that handles fetching settings values.
- `app_settings`: An instance of the `MimicrySettings` class which is used to access the settings.

Default Settings:
- `MIMICRY_FEATURE_CONTROL_CONDITION`: Default value is 'is_staff'. Determines the attribute or condition to control the
Mimicry functionality.
- `PRIVATE_CONTENT_REPLACEMENT`: Default value is 'HIDDEN FOR USER PRIVACY PURPOSES'. The replacement content for
private or sensitive information during simulation.
- `SIMULATED_SESSION_EXPIRY`: Default value is 3600. The time (in seconds) for how long a simulated session lasts before
expiry.
- `SIMULATED_SESSION_RETENTION`: Default value is 14. The retention period (in days) for storing simulated session data.
- `ONLY_ALLOW_SIMULATED_GET_REQUESTS`: Default value is True. Restricts simulated user sessions to only allow GET and
HEAD requests.
- `ENABLE_MIMICRY`: Default value is True. Flag to enable or disable the entire Mimicry functionality.
- `MIMICRY_AUTHENTICATION_BACKEND`: Default value is 'django.contrib.auth.backends.ModelBackend'. The
authentication backend used for simulated users.

Usage:
Import the `app_settings` object and access the settings attributes as needed.
"""

from django.conf import settings
from django.contrib.auth import get_user_model


UserModel = get_user_model()


class SimulateUserSettings:
    """
    Handles default settings and project overridden settings for the Mimicry app.
    """

    # Define default settings
    defaults: dict[str, any] = {
        'MIMICRY_FEATURE_CONTROL_CONDITION': 'is_staff',
        'PRIVATE_CONTENT_REPLACEMENT': 'HIDDEN FOR USER PRIVACY PURPOSES',
        'SIMULATED_SESSION_EXPIRY': 3600,
        'SIMULATED_SESSION_RETENTION': 14,
        'ONLY_ALLOW_SIMULATED_GET_AND_HEAD_REQUESTS': True,
        'ENABLE_MIMICRY': True,
        'ENABLE_SIMULATE_USER_NOTIFICATIONS': True,
        'MIMICRY_AUTHENTICATION_BACKEND': 'django.contrib.auth.backends.ModelBackend',
        'MIMICRY_PERMISSIONS': [
            {"SIMULATED_USER_ATTRIBUTE": "is_active", "REAL_USER_ATTRIBUTE": "is_staff"}
        ]
    }

    expected_types: dict[str, any] = {
        'MIMICRY_FEATURE_CONTROL_CONDITION': str,
        'PRIVATE_CONTENT_REPLACEMENT': str,
        'SIMULATED_SESSION_EXPIRY': int,
        'SIMULATED_SESSION_RETENTION': int,
        'ONLY_ALLOW_SIMULATED_GET_AND_HEAD_REQUESTS': bool,
        'ENABLE_MIMICRY': bool,
        'ENABLE_SIMULATE_USER_NOTIFICATIONS': bool,
        'MIMICRY_AUTHENTICATION_BACKEND': str,
        'MIMICRY_PERMISSIONS': list
    }

    def __getattr__(self, name: str) -> any:
        value = getattr(settings, name, self.defaults.get(name))
        expected_type = self.expected_types.get(name)

        if expected_type and not isinstance(value, expected_type):
            raise TypeError(f"Expected type for {name} is {expected_type}, but got {type(value)}.")

        if name == 'MIMICRY_FEATURE_CONTROL_CONDITION':
            self.validate_user_control_condition(value)

        if name == 'MIMICRY_PERMISSIONS':
            self.validate_user_permissions(value)

        return value

    @staticmethod
    def validate_user_control_condition(condition: str) -> None:
        """Validates if the given condition corresponds to a boolean attribute of the User model."""
        if not hasattr(UserModel, condition):
            raise ValueError(f"'{condition}' is not an attribute of the User model.")

        attribute_type = type(getattr(UserModel, condition))
        if attribute_type != bool:
            raise ValueError(f"'{condition}' is not a boolean attribute of the User model.")

    def validate_user_permissions(self, permissions: list) -> None:
        """Validates the SIMULATE_USER_PERMISSIONS setting."""
        if not all(isinstance(perm, dict) for perm in permissions):
            raise TypeError("All elements in 'MIMICRY_PERMISSIONS' should be dictionaries.")

        for perm in permissions:
            for key in ['SIMULATED_USER_ATTRIBUTE', 'REAL_USER_ATTRIBUTE']:
                if key not in perm:
                    raise TypeError(
                        f"Each dictionary in 'MIMICRY_PERMISSIONS' must contain the key '{key}'.")

                self.validate_user_control_condition(perm[key])


# Create an instance of the class, so we can use its attributes
app_settings: SimulateUserSettings = SimulateUserSettings()
